/*;  ltank.h {\documentstyle[fleqn]{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.1 $$ {\rhead{\sl ltank.h\quad\thepage}} */
/*;  $$Date: 2000/10/14 15:10:07 $$  */
/*;  $$Source: /home/shiiba/ohymos-lib/ohymos-current/doc/element/ltank.h,v $$  */
/*;  $$RCShost: yodo.kuciv.kyoto-u.ac.jp $$  */

// @(#) header file for class KinemKB

#if !defined(_ltank_h)
#define _ltank_h

#include <stdio.h>
#include <time.h>

#include "element.h"   // Definition of class Element
#include "p_tmp.h"     // Definition of template classes Receive
                       // and Send
extern "C"{
#include <math.h>      // to use exp
}

//  Definitio of class Ltank

/*----------------------------------------------------------------------
CLASS

  Ltank

  Linear storage model

KEYWORD

  tank storage linear

DESCRIPTION

  Ltank is a linear storage model in which the time chage of s(t) the
  storage at time t is described by the orfinary differential equation

  ds/dt = I(t) - a s(t) - b s(t)

  where I(t) is the input at time t, O_l(t) = a s(t) and O_b(t) = b s(t)
  are outputs.

  The solution is given by

  <pre>
  s(t + \Delta t) = s(t) \exp(-(a+b)\Delta t) +
              \int_t^(t+\Delta t) I(u) \exp\{(a+b)(u - t)\} du
  </pre>
  Setting \lambda = (a + b) \Delta t and assuming that I(t) takes a
  constant value I_m from t to t + \Delta t, we get

  s(t + \Delta t) = c_1 s(t) + c_2 I_m

  where

  c_1 = exp(-\lambda), c_2 = (1 - c_1) / (a + b)

  [Receive ports]<br>
  One receive port of type Receive<double> with tne object name "rp" and
  the object number 0. I(t) = mC * the value obtained through this
  receive port.  The data connection type of the data can be selected
  (See dct in the Parameter section).

  [Send ports]<br>
  Two receive ports.

  The send port of type Send<double> with the object name "sp" and the
  object number 0 sends O_l(t) outflow from the side outflow hole. The
  data connection type of the send port is broken line.

  The send port of type Send<double> with the object name "sp" and the
  object number 1 sends O_b(t) outflow from the bottom outflow hole. The
  data connection type of the send port is broken line.

  [Parameters]<br>
  <pre>
    mA   ..... a (1/sec)
    mB   ..... b (1/sec)
    mLambda .. lambda (-)
    mC   ..... magnification to the value obtained through the receive
               port ("rp", 0) to get I(t).
    dct  ..... data connection type of the receive port ("rp", 0).
               0 ... broken line
               1 ... step
               2 ... pulse
  </pre>

  [State variables]<br>
  storage s(t) and time

  [Note]<br>
  s(t) can be regarded as either the volume or the depth. When s(t) is
  volume, the dimension of I(t) = mC * the value obtained through the
  receive port ("rp", 0) and the dimension of outflow are volume per
  sec. When s(t) is depth, the dimensions of I(t) and outflow are depth
  per sec.

----------------------------------------------------------------------*/

class Ltank : public Element {      // (1) Inherit class Element

//  1.  Definition of receive and send ports

  // Receive ports
protected:

  // (2) Declare receive ports as data members

  //////////////////////////////////////////////////////////////////////
  // receive port ("rp", 0) which receives I(t).  See mC also.
  // I(t) = mC * value obtained through the receive port ("rp", 0).
  Receive<double> mRp;

  // (3) Define void Register_receive_ports(void) which registers
  //     receive ports

  //////////////////////////////////////////////////////////////////////
  // register receive ports
  void Register_receive_ports(void)
  {
    Register((void*) &mRp);
  }

  // Send ports
protected:

  // (4) Declare send ports as data members

  //////////////////////////////////////////////////////////////////////
  // mSp0: send port ("sp", 0) which outputs outflow intensity from a
  // side hole
  Send<double> mSp0;

  //////////////////////////////////////////////////////////////////////
  // mSp1: send port ("sp", 1) which outputs outflow intensity form the
  // bottom
  Send<double> mSp1;

  // (5) Define void Register_send_ports(void) which regisers
  //     send ports

  //////////////////////////////////////////////////////////////////////
  // register send ports
  void Register_send_ports(void)
  {
    Register((void*) &mSp0);
    Register((void*) &mSp1);
  }

//  2.  Parameters and state variables

// parameters

  // (6) Declares parameters as data members

protected:

  //////////////////////////////////////////////////////////////////////
  // magnification mC.  I(t) = mC *  value obtained through the receive
  // port ("rp", 0).
  double mC;

  //////////////////////////////////////////////////////////////////////
  // parameter a
  double mA;

  //////////////////////////////////////////////////////////////////////
  // parameter b
  double mB;

  //////////////////////////////////////////////////////////////////////
  // lambda = (a + b) $\Delta$ t, where $\Delta$ t is time increment
  double mLambda;

// state varibales

  // (7) Declare state variables as data members

protected:
  //////////////////////////////////////////////////////////////////////
  // s(t) the storarge at time t
  double mS;                   // storage $s(t)$

// working variables

  // (8) Delare working variables if needed

protected:
  double mC1, mC2;     // parameter $c_1$, $c_2$
  time_t mDt;          // time increment $\Delta t$
  double mY;           // inflow intensity at time t I(t)

  // Set variables to paramers and initial state

public:

  // (9) Define void Set_parameter(FILE*  fp) which reads in
  //     parameter values from a file

  //////////////////////////////////////////////////////////////////////
  // set parameter values from a file.  Parameter values are read by
  // <pre>
  //  Get_line(fp, bf, BUFFLEN, __FILE__, __LINE__);
  //  sscanf(bf, "%lf %lf %lf %lf %d", &mA, &mB, &mLambda, &mC, &dct);
  // </pre>
  // where
  // <pre>
  //  mA   ..... a (1/sec)
  //  mB   ..... b (1/sec)
  //  mLambda .. lambda (-)
  //  mC   ..... magnification to the value obtained through the receive
  //             port ("rp", 0) to get I(t).
  //  dct  ..... data connection type of the receive port ("rp", 0).
  //             0 ... broken line
  //             1 ... step
  //             2 ... pulse
  // </pre>
  void Set_parameter(FILE* fp)
  {
    const int BUFFLEN = 256;   char bf[BUFFLEN];

    Get_line(fp, bf, BUFFLEN, __FILE__, __LINE__);
    // skip lines which start wid # and while lines

    int dct;
    sscanf(bf, "%lf %lf %lf %lf %d", &mA, &mB, &mLambda, &mC, &dct);
    mRp.Set_data_connection_type(dct);
    mDt = (time_t)(mLambda/ (mA + mB));
    mLambda = mDt * (mA + mB); // $\lambda = \Delta t /(a + b)$
    mC1 = exp(-mLambda);            // $c_1 = \exp\{-\lambda\}$
    mC2 = (1.0 - mC1)/(mA + mB);
    // $c_2 = (1 - c_1) / (a + b)$
  }

  // Initillize state variables
public:

  // (10) Define void Set_initial_state(FILE* fp)
  //      which reads initial state values from a file

  //////////////////////////////////////////////////////////////////////
  // set initial state by
  // <pre>
  // sscanf(bf, "%ld %lf", &current_time, &mS);
  // </pre>
  void Set_initial_state(FILE* fp)
  {
    const int BUFFLEN = 256;   char bf[BUFFLEN];
    Get_line(fp, bf, BUFFLEN, __FILE__, __LINE__);
    // skip lines which start wid # and while lines
    sscanf(bf, "%ld %lf", &current_time, &mS);
  }

  // Save state variables
public:

  // (11) Define void Save_terminal_state(FILE* fp)
  //      which saves the terminal state into a file

  //////////////////////////////////////////////////////////////////////
  // save terminal state into a file by
  // <pre>
  // fprintf(fp, "%ld %lf", current_time, mS);
  // </pre>
  void Save_terminal_state(FILE* fp);

//  3.  Init, constructor, descructior

public:

  // (12) Define NewElement()

  //////////////////////////////////////////////////////////////////////
  // define NewElement which returns a pointer to a new Ltank created in
  // heap
  Ltank* NewElement() { return new Ltank(); }

private:

  //; (12) Only declare operator =, but do not give its definition

  Ltank& operator=(const Ltank& rhs);

public:

  //////////////////////////////////////////////////////////////////////
  // Init function for the Structure Configuration File method.
  Boolean Init(char* objname, int obj_num, FILE* fp) {
    Boolean reply = Init(objname, obj_num);
    if (reply == YES) Set_parameter(fp);
    return reply;
  }

  //////////////////////////////////////////////////////////////////////
  // Initialize the memory
  Boolean Init(char* o_name, int o_num, char* p_str = NULL)
        { return Element::Init(o_name, o_num, 0, p_str); }

  //; (13) Constructor

  //////////////////////////////////////////////////////////////////////
  // Constructor
  Ltank(char* o_name, int o_num, char* p_str = NULL)
        :   Element(),
            mRp("rp", 0),
            mSp0("sp", 0),
            mSp1("sp", 1)
        { class_name = "Ltank"; Init(o_name, o_num, p_str); }
    Ltank(void)
        :   Element(),
            mRp("rp", 0),      //; Receive port ("rp", 0)
            mSp0("sp", 0),     //; Send port ("sp", 0)
            mSp1("sp", 1)      //; Send port ("sp", 1)
        { class_name = "Ltank"; }
    ~Ltank(void) {}

//  4.  Calcuration

  // (14) Define Boolean Initial_output(void) which outputs
  //      initial outputs.
  //      Output unit of output data by the use of Set_comment_string

  //////////////////////////////////////////////////////////////////////
  // Initial output function
  Boolean Initial_output(void) {
    if (mRp.Can_you_get_data(current_time) == NO) return NO;
    mY = mRp.Get_data(current_time);
    double data = mA * mS;
    mSp0.Set_comment_string("time(sec) discharge height(m/sec)");
    mSp0.Send_data(current_time, data);
    data = mB  * mS;
    mSp1.Set_comment_string("time(sec) storage depth(m)");
    mSp1.Send_data(current_time, data);
    // printf("Ltank::Initial_output done\n");
    return YES;
  }

  // (15) Define time_t Calculate_time_step(void) which returns
  //      the time step

  //////////////////////////////////////////////////////////////////////
  // Calculate time step. Simply returns mDt = lambda /(a + b)
  time_t Calculate_time_step(void) { return mDt; }

  // (16) Define Boolean Can_you_calculate(void)

  //////////////////////////////////////////////////////////////////////
  // Can_you_calculate function. If input data from cuttent_time to
  // current_time + mDt can be got, it returns YES.
  Boolean Can_you_calculate(void) { return Can_you_calculate0(); }

  // (17) Define Boolean Calculate(void)

  //////////////////////////////////////////////////////////////////////
  // Calculate s(t + mDt) and returns outputs at t + mDt
  Boolean Calculate(void);

  // (18) Define Boolean Work(void)

  //////////////////////////////////////////////////////////////////////
  // Work. Simply Work0 is called. From within Work0,
  // Calculate_time_step() and Can_you_calculate() are called.
  Boolean Work(void) { return Work0(); }
};

#endif  //  ! _ltank_h

/*; Local Variables: *** */
/*; tab-width: 4 *** */
/*; fill-column: 72 *** */
/*; End: *** */

/*; end of file */
